#Arguments for the command:

#ARGV[0] = number of generations
#ARGV[1] = population size
#ARGV[2] = mutation rate
#ARGV[3] = name of the folder to store the results (relative to C:\\Genetic_Experiments)
#ARGV[4] = 0 if the fitness is binary, 1 if is a number

#example: 
#ruby genetic-command.rb 200 100 0.1 Test1 0

#class Numeric of Ruby
class Numeric
	#returns:: this number
	#This method is defined in the API of SketchUp in order to work with meters instead of inches. When this command is used, we are working without SkethUp,
	#so we do not need this conversion. We define thus this method for avoiding errors when the invocations of .m are done; since Ruby does not define it.
	#Obviously, when SketchUp is not being used, this method does not do anything, returning the same number.
	def m
		return self
	end
end

require "#{File.dirname(__FILE__)}/../main-structures"
require "#{File.dirname(__FILE__)}/../utils"

Shade.using_sketchup = false
#Create the project
ShadeUtils.create_default_project

#Set list of constraint names and classes
ShadeUtils.add_cg_names()

#Search for custom constraints
ShadeUtils.load_custom_constraints()

#Search for custom goals
ShadeUtils.load_custom_goals()

GENOTYPE_LENGTH = 28

def get_phenotype(genotype)
	
	Shade.project.execution.reset
	i =0
	while (i < GENOTYPE_LENGTH)
		
		bit3 = genotype[i]
		bit2 = genotype[i+1]
		bit1 = genotype[i+2]
		bit0 = genotype[i+3]
		
		rule_number = bit0 + (bit1*2) + (bit2*4) + (bit3*8)
		
		t = get_transformation(Shade.project.execution.current_shape, rule_number)
		
		Shade.project.execution.apply_rule(1, t)
		
		i += 4
	end
	
	return Shade.project.execution.current_shape
end

def get_transformation(shape, rule_number)
	
	#First, we make the list of points representing the center labels. For every point in this list, there is a square in the shape
	label_points = shape.p["Layer0"].get_node(Label.new("Black")).list
	aux_labels = label_points.clone
	
	#Second, we count the number of possible rule applications
	n_possible_rules = 0
	label_points.reset_iterator
	while (p = label_points.get_next)
		
		
		top_neighbor = OrderedPoint.new(Point.new(p.key.x, p.key.y + 6, 0))
		right_neighbor = OrderedPoint.new(Point.new(p.key.x + 6, p.key.y, 0))
		bottom_neighbor = OrderedPoint.new(Point.new(p.key.x, p.key.y - 6, 0))
		left_neighbor = OrderedPoint.new(Point.new(p.key.x - 6, p.key.y, 0))
		if !aux_labels.get_node top_neighbor
			n_possible_rules += 1
		end
		if !(aux_labels.get_node right_neighbor)
			n_possible_rules += 1
		end
		if !(aux_labels.get_node bottom_neighbor)
			n_possible_rules += 1
		end
		if !(aux_labels.get_node left_neighbor)
			n_possible_rules += 1
		end
	end
	
	rule_number = (rule_number % n_possible_rules)
	
	label_points.reset_iterator
	
	index = -1
	found = false
	while ((p = label_points.get_next) && !found)
		
		
		x_origin = p.key.x - 3
		y_origin = p.key.y - 3
		top_neighbor = OrderedPoint.new(Point.new(p.key.x, p.key.y + 6, 0))
		right_neighbor = OrderedPoint.new(Point.new(p.key.x + 6, p.key.y, 0))
		bottom_neighbor = OrderedPoint.new(Point.new(p.key.x, p.key.y - 6, 0))
		left_neighbor = OrderedPoint.new(Point.new(p.key.x - 6, p.key.y, 0))
		if !aux_labels.get_node top_neighbor
			index += 1
			if index == rule_number
				found = true
				transformation = [0, -1, x_origin + 6, 1, 0, y_origin]
			end
		end
		if (!found && !(aux_labels.get_node right_neighbor))
			index += 1
			if index == rule_number
				found = true
				transformation = [1, 0, x_origin , 0, 1, y_origin]
			end
		end
		if (!found && !(aux_labels.get_node bottom_neighbor))
			index += 1
			if index == rule_number
				found = true
				transformation = [0, 1, x_origin, -1, 0, y_origin + 6]
			end
		end
		if (!found && !(aux_labels.get_node left_neighbor))
			index += 1
			if index == rule_number
				found = true
				transformation = [-1, 0, x_origin + 6, 0, 1, y_origin]
			end
		end
	end
	
	return transformation
end

def binary_fitness(genotype)
	
	shape = get_phenotype(genotype)
	
	label_points = shape.p["Layer0"].get_node(Label.new("Black")).list
	label_points.reset_iterator
	i = 0
	x_labels = Array.new
	while  (p = label_points.get_next)
		if (i == 0)
			x_labels.push p.key.x
		else
			if !(p.key.x == x_labels.last)
				x_labels.push p.key.x
			end
		end
		i+=1
	end
	
	x_size = x_labels.size
	middle_x = 0
	if ((x_size % 2) == 0)
		middle_pos = (x_size/2)-1
		middle_x = x_labels[middle_pos] + 3
	else
		middle_pos = (x_size/2)
		middle_x = x_labels[middle_pos]
	end
	
	aux_points = label_points.clone
	label_points.reset_iterator
	coincidences = 0
	while (p = label_points.get_next)
		if p.key.x < middle_x
			delta_x = middle_x - p.key.x
			sym_square = OrderedPoint.new(Point.new(middle_x + delta_x, p.key.y, 0))
			result = aux_points.get_node(sym_square)
			#UI.messagebox("(#{sym_square.x.to_m}, #{sym_square.y})?, #{result}")
		elsif p.key.x > middle_x
			delta_x = p.key.x - middle_x
			sym_square = OrderedPoint.new(Point.new(middle_x - delta_x, p.key.y, 0))
			result = aux_points.get_node(sym_square)
			#UI.messagebox("(#{sym_square.x.to_m}, #{sym_square.y})?, #{result}")
		elsif p.key.x == middle_x
			result = true #The points which are in the vertical axis do sum reward
		end
		if result
			coincidences += 1
		end
	end

	result = 0
	if coincidences == 8
		result = 1
	end
	return result
end

def fitness(genotype)
	
	shape = get_phenotype(genotype)
	
	label_points = shape.p["Layer0"].get_node(Label.new("Black")).list
	label_points.reset_iterator
	i = 0
	x_axis = Array.new
	while  (p = label_points.get_next)
		if (i == 0)
			x_axis.push p.key.x
			x_axis.push p.key.x + 3
		else
			last = x_axis.size - 2
			if !(p.key.x == x_axis[last])
				x_axis.push p.key.x
				x_axis.push p.key.x + 3
			end
		end
		i+=1
	end
	
	x_axis.pop
	
	x_size = x_axis.size
	i = 0
	min_dif = nil
	while (i < x_size)
		
		n_x_less = 0
		n_x_greater = 0
		
		label_points.reset_iterator
		while  (p = label_points.get_next)
			if p.key.x < x_axis[i]
				n_x_less += 1
			elsif p.key.x > x_axis[i]
				n_x_greater += 1
			end
		end

		dif = (n_x_less - n_x_greater).abs
		
		if !min_dif
			min_dif = dif
			middle_x = x_axis[i]
		else
			if dif < min_dif
				min_dif = dif
				middle_x = x_axis[i]
			end
		end
		i+=1
	end
	
	aux_points = label_points.clone
	label_points.reset_iterator
	coincidences = 0
	while (p = label_points.get_next)
		if p.key.x < middle_x
			delta_x = middle_x - p.key.x
			sym_square = OrderedPoint.new(Point.new(middle_x + delta_x, p.key.y, 0))
			result = aux_points.get_node(sym_square)
			#UI.messagebox("(#{sym_square.x.to_m}, #{sym_square.y})?, #{result}")
		elsif p.key.x > middle_x
			delta_x = p.key.x - middle_x
			sym_square = OrderedPoint.new(Point.new(middle_x - delta_x, p.key.y, 0))
			result = aux_points.get_node(sym_square)
			#UI.messagebox("(#{sym_square.x.to_m}, #{sym_square.y})?, #{result}")
		elsif p.key.x == middle_x
			result = true #The points which are in the vertical axis do sum reward
		end
		if result
			coincidences += 1
		end
	end

	return coincidences
end


def mutate(phenotype, mutation_rate)
	position = rand(GENOTYPE_LENGTH)
	
	mutated_phenotype = phenotype
	
	if rand < mutation_rate
		if phenotype[position] == 1
			mutated_phenotype[position] = 0
		else
			mutated_phenotype[position] = 1
		end
	end
	
	return mutated_phenotype
end

def recombine(father, mother)
	child1 = Array.new(GENOTYPE_LENGTH)
	child2 = Array.new(GENOTYPE_LENGTH)
	
	crossover_position = rand(GENOTYPE_LENGTH-1)
	
	child1 = father[0..crossover_position] + mother[(crossover_position + 1)..GENOTYPE_LENGTH-1]
	child2 = mother[0..crossover_position] + father[(crossover_position + 1)..GENOTYPE_LENGTH-1]
	
	return [child1, child2]
end

def get_initial_population(population_size)
	population = Array.new
	
	population_size.times do
		individual = Array.new
		
		GENOTYPE_LENGTH.times do 
			individual.push rand(2)
		end
		
		population.push individual
	end
	
	return population
end

def get_next_generation(population)
	selected_individuals = get_selected_individuals(population)
	
	next_generation = Array.new
	
	while !selected_individuals.empty?
		father_pos = rand(selected_individuals.size)
		father = selected_individuals[father_pos]
		selected_individuals.delete_at(father_pos)
		
		mother_pos = rand(selected_individuals.size)
		mother = selected_individuals[mother_pos]
		selected_individuals.delete_at(mother_pos)
		
		children= recombine(father, mother)
		
		next_generation.push mutate(children[0], $mutation_rate)
		next_generation.push mutate(children[1], $mutation_rate)
	end
	
	return next_generation
end

def get_selected_individuals(population)
	
	total_fitness = 0
	
	population.each { |individual|
		if $binary_fitness
			fitness = binary_fitness(individual)
		else
			fitness = fitness(individual)
		end
		
		total_fitness += fitness
	}
	
	individual_probabilities = Array.new
	last = nil
	
	if total_fitness > 0
		population.each {|individual|
			if $binary_fitness
				fitness = binary_fitness(individual)
			else
				fitness = fitness(individual)
			end
			
			probability = fitness.quo(total_fitness)
			if last
				inf = last
				sup = probability + last
				individual_probabilities.push [individual, [inf, sup]]
			else
				inf = 0
				sup = probability
				individual_probabilities.push [individual, [inf, sup]]
			end
			last = sup
		}
		
		selected = Array.new
		
		while (selected.size < population.size)
			
			roulette = rand
			found = false
			i = 0
			while (!found && (i < individual_probabilities.size))
				if ((roulette >= individual_probabilities[i][1][0]) &&(roulette < individual_probabilities[i][1][1]))
					found = true
					selected.push individual_probabilities[i][0]
				end
				i+=1
			end
		
		end

	else
		selected = population
	end
	
	return selected
	
end

def save_generation(population, directory)
	i = 0
	population.each { |individual|
		shape = get_phenotype(individual)
		shape.save("#{directory}\\shape#{i}.txt")
		
		j =0
		individual_dec = Array.new
		while (j < GENOTYPE_LENGTH)
			
			bit3 = individual[j]
			bit2 = individual[j+1]
			bit1 = individual[j+2]
			bit0 = individual[j+3]
			
			rule_number = bit0 + (bit1*2) + (bit2*4) + (bit3*8)
			
			individual_dec.push rule_number
			
			j += 4
		end
		
		File.open("#{directory}\\genotype#{i}.txt", 'w') do |f|
			f.write "#{individual}\n"
			f.write "#{individual_dec}\n"
		end
		i+=1
	}
end


$n_generations = ARGV[0].to_i
$population_size = ARGV[1].to_i
$mutation_rate = ARGV[2].to_f
$binary_fitness = false
if ARGV[4].to_i == 0
	$binary_fitness = true
end

experiment_directory = "C:\\Genetic_Experiments\\#{ARGV[3]}"
Dir.mkdir experiment_directory unless File.directory?(experiment_directory)

File.open("#{experiment_directory}\\data.txt", 'w') do |f|
	f.write "Number of generatons: #{$n_generations}\n"
	f.write "Population_size: #{$population_size}\n"
	f.write "Mutation rate: #{$mutation_rate}\n"
	f.write "Binary fitness?: #{$binary_fitness}\n"
end

grammar_path = "C:\\Symmetry\\sym_weak.gr2"
Shade.project.execution.grammar.load(grammar_path)

population = get_initial_population($population_size)
Dir.mkdir("#{experiment_directory}\\Generation0") unless File.directory?("#{experiment_directory}\\Generation0")
save_generation(population, "#{experiment_directory}\\Generation0")

i = 1
$n_generations.times do 
	population = get_next_generation(population)
	Dir.mkdir("#{experiment_directory}\\Generation#{i}") unless File.directory?("#{experiment_directory}\\Generation#{i}")
	save_generation(population, "#{experiment_directory}\\Generation#{i}")
	i+=1
end

puts "End"
