require 'sketchup' # First we pull in the standard API hooks.
require 'ShaDe//lib//main-structures.rb'
require 'ShaDe//lib//utils.rb'
require 'ShaDe//lib//interfaces//guitools.rb'
require 'ShaDe//lib//layers.rb'

# Show the Ruby Console at startup so we can
# see any programming errors we may make.
Sketchup.send_action "showRubyPanel:"

# MAIN CODE ################################################################
# Add a menu item to launch our plugin, and prepare the plugin variables
menu = UI::menu("Plugins");
menu.add_separator;
menu.add_item("ShaDe") {

	#We open a new file
	Sketchup.file_new
	
	#We are inside SketchUp
	Shade.using_sketchup = true
	
	#We set the size of labels
	Shade.label_radius = Constants::LABEL_RADIUS
	
	#We prepare the SU canvas
	ShadeUtils.prepare_canvas
	
	#Create the default project
	ShadeUtils.create_default_project
	
	#Create the toolbars
	create_static_toolbar()
	create_execution_toolbar()
	create_edition_toolbar()
	create_guidance_toolbar()
	
	#Add the close observer
	close_observer = CloseObserver.new
	Sketchup.add_observer(close_observer)
	
	#Add the undo observer
	undo_observer = UndoObserver.new
	Sketchup.active_model.add_observer(undo_observer)
	
	#Add the layers observer
	layers_observer = ShaDeLayersObserver.new
	Sketchup.active_model.layers.add_observer(layers_observer)
	
	#Refresh the project
	Shade.project.refresh(true)
	
	#Explicitly say that the grammar is saved
	Shade.project.execution.grammar.saved = true
	
	#Set showing text to false
	Shade.show_text = false
	
	#Set scripting robot to nil
	Shade.robot = nil
	
	#Set list of constraint names and classes
	ShadeUtils.add_cg_names()
	
	#Search for custom constraints
	ShadeUtils.load_custom_constraints()
	
	#Search for custom goals
	ShadeUtils.load_custom_goals()
}